/*
   File       : oqs_cre_procedure.sql
	 Author     : Verweyen Consulting
	              Georg Verweyen
	 History    : 20070406 Creation
	 Parameters : n/a
	 Tables     : n/a
	 Procedure  : oqs_working       Creation
	              Parameters : &1.  table owner
								             &2.  table name
														 &3   group value
	              Tables     : oqs_meta_data   Select
	                           oqs_kpi_type    Select
														 <&1..&2.>       Select (full table scan)
	              						 oqs_kpi_log_run Insert
	 Description: Creates the PL/SQL procedure oqs_working
	              The procedure generate a dynamical SQL statement
								for the attributes of the presented table (like min, max, ..)
								
	              The number of parallel threads is set on 1!
*/
CREATE OR REPLACE PROCEDURE oqs_working (
  table_owner VARCHAR2,
  table_name  VARCHAR2,
	group_value VARCHAR2
) IS
  CURSOR_CLOSED       EXCEPTION;
	STATEMENT_TOO_LONG  EXCEPTION;
  PRAGMA EXCEPTION_INIT (CURSOR_CLOSED, -1002);
  CURSOR c_qswork(v_owner IN VARCHAR, v_table_name IN VARCHAR) IS 
    SELECT m.column_name, m.column_len, m.column_type,
		       m.kpi_name,  m.kpi_group,
           m.kpiwh_name, m.kpiwh_type, m.kpiwh_value,
					 t.result_type, t.statement_1, t.statement_2,
					 t.null_count
			FROM oqs_meta_data m, oqs_kpi_type t
		 WHERE m.owner      = v_owner
		   AND m.table_name = v_table_name
			 AND m.kpi_type   = t.kpi_type
			 AND m.column_type= t.column_type
		 ORDER BY m.column_name, m.kpi_type
  ;
	v_column_name  oqs_meta_data.column_name%TYPE;
	v_column_len   oqs_meta_data.column_len%TYPE;
	v_column_type  oqs_meta_data.column_type%TYPE;
	v_kpi_name     oqs_meta_data.kpi_name%TYPE;
  v_kpi_group    oqs_meta_data.kpi_group%TYPE;
  v_kpiwh_name   oqs_meta_data.kpiwh_name%TYPE;
  v_kpiwh_type   oqs_meta_data.kpiwh_type%TYPE;
  v_kpiwh_value  oqs_meta_data.kpiwh_value%TYPE;
  v_res_type     oqs_kpi_type.column_type%TYPE;
	v_wrk_st1      oqs_kpi_type.statement_1%TYPE;
	v_wrk_st2      oqs_kpi_type.statement_2%TYPE;
	v_null_count   oqs_kpi_type.null_count%TYPE;
	v_starttime    date;
  c_stat    NUMBER;
  v_ignore  NUMBER;
  cnt       PLS_INTEGER;
	v_cnt     NUMBER;
  v_stat    varchar2(32767); -- Enthlt das Statement
	v_st_part varchar2(32767);
  v_number  NUMBER;          -- Ergebnis einer NUMBER-COLUMN
  v_string  varchar2(2000);  -- Ergebnis einer VARCHAR-COLUMN (2000 kann auf max. Lnge angepasst werden)
	v_work    varchar2(100);
  --                        -- Typen fr PL/SQL-Tabellen
  TYPE t_number  IS TABLE OF NUMBER INDEX BY BINARY_INTEGER;
  TYPE t_cattype IS TABLE OF VARCHAR2(1) INDEX BY BINARY_INTEGER;
  TYPE t_catname IS TABLE OF VARCHAR2(70) INDEX BY BINARY_INTEGER;
  TYPE t_stat_s  IS TABLE OF VARCHAR2(255) INDEX BY BINARY_INTEGER;
  v_coltype t_cattype;                    -- Typ der Column 1=VARCHAR, 2=NUMBER
  v_collen  t_number;                     -- Lnge der Column (bei VARCHAR)
  v_colname t_catname;                    -- Name der Spalte (Achtung CaseSensitiv)
  v_colcnt  number;                       -- Anzahl der Spalten
  v_stat_s  sys.dbms_sql.varchar2s;       -- Statement in 255 Byte-Blcken
	

BEGIN
  -- Number of parallel threads can changed
	v_stat   := 'SELECT /*+ PARALLEL (w,1) */'; -- the number (and only this) can changed
	-- End of changing sector
	v_colcnt := 0;
	v_starttime := SYSDATE;
	OPEN c_qswork(table_owner, table_name);
	LOOP 
	  FETCH c_qswork INTO 
		  v_column_name, v_column_len,
			v_column_type, v_kpi_name,
		  v_kpi_group, v_kpiwh_name, 
      v_kpiwh_type, v_kpiwh_value,
      v_res_type, v_wrk_st1, 
			v_wrk_st2, v_null_count;
		EXIT WHEN c_qswork%NOTFOUND;
		v_colcnt := v_colcnt+1;
 	  v_coltype(v_colcnt) := v_res_type;
    v_collen (v_colcnt) := v_column_len ;
    v_colname(v_colcnt) := v_kpi_name;
	  if (v_colcnt > 1) then
	    v_st_part := ', ';
		ELSE
	    v_st_part := '';
	  end if;
		/*
		   count(distinct case NR when 2 then TEST else NULL end)
		*/
 	  v_st_part := v_st_part||v_wrk_st1;
		if (v_kpiwh_name is not NULL) then
		  v_st_part:=v_st_part||'case '||v_kpiwh_name||' when ';
			IF (v_kpiwh_type='V') THEN
  			v_st_part:=v_st_part||'''';
			END IF;
			v_st_part:=v_st_part||v_kpiwh_value;
			IF (v_kpiwh_type='V') THEN
  			v_st_part:=v_st_part||'''';
			END IF;
			v_st_part:=v_st_part||' then ';
		end if;
		v_st_part := v_st_part||v_column_name;
		if (v_kpiwh_name is not NULL) then
			v_st_part:=v_st_part||' ELSE';
			IF (v_null_count='N') THEN
  			v_st_part:=v_st_part||' NULL';
			ELSE
			  IF (v_column_type='V') THEN
    			v_st_part:=v_st_part||'''';
			  END IF;
  			v_st_part:=v_st_part||' 1';
			  IF (v_column_type='V') THEN
    			v_st_part:=v_st_part||'''';
			  END IF;
			END IF;
			v_st_part:=v_st_part||' END';
		end if;
		v_st_part := v_st_part||v_wrk_st2;
		--dbms_output.put_line (v_kpi_name || ': '||length(v_st_part));
		IF length(v_stat)+length(v_st_part) > 32767 THEN
		  RAISE STATEMENT_TOO_LONG;
		END IF;
		v_stat := v_stat||v_st_part;
	END LOOP;
	IF (v_colcnt=0) THEN
		RAISE NO_DATA_FOUND;
	END IF;
	v_stat := v_stat || ' FROM '||table_owner||'.'||table_name|| ' w ';
	if (group_value>' ') then
	  v_stat := v_stat || ' WHERE '||v_kpi_group||'='||group_value;
		--dbms_output.put_line (' WHERE '||v_kpi_group||'='||group_value);
	end if;
  c_stat   := sys.dbms_sql.open_cursor;
	/*
	procedure parse(c in integer, statement in varchar2s,
                  lb in integer, ub in integer,
                  lfflg in boolean, language_flag in integer);
    create_stmt(1):= 'create table '|| TABLENAME;
    create_stmt(2) := '(id number, name varchar2(20))';
    cursor_id := dbms_sql.open_cursor;
    dbms_sql.parse(cursor_id, create_stmt, 1, 2, NULL , dbms_sql.native);
	*/
	cnt := 1;
	-- print the first 2000 charactes of the statement
	--dbms_output.put_line (substr(v_stat,1,2000));
	while length(v_stat)>255 LOOP
	  v_stat_s(cnt) := substr(v_stat,1,255);
		v_stat        := substr(v_stat,256);
		cnt := cnt +1;
	END LOOP;
	v_stat_s(cnt) := v_stat;
	-- technical information 
	--dbms_output.put_line (cnt||' Teilstuecke');
  sys.dbms_sql.parse(c_stat,v_stat_s,1,cnt,NULL, sys.dbms_sql.native);
  v_ignore := sys.dbms_sql.execute(c_stat);
	--dbms_output.put_line (v_colcnt); 
  for v_cnt in 1..v_colcnt
	LOOP
  	--dbms_output.put (v_cnt); 
  	IF (v_coltype(v_cnt)='V') THEN
    	--dbms_output.put_line (v_cnt||' String '||v_colname(v_cnt)||' - '||v_coltype(v_cnt)||' - '||v_collen (v_cnt)); 
      SYS.DBMS_SQL.DEFINE_COLUMN (c_stat, v_cnt, v_string, v_collen(v_cnt));
	  END IF;
	  IF (v_coltype(v_cnt)='N') THEN
    	--dbms_output.put_line (' Zahl '||v_colname(v_cnt)); 
     	SYS.DBMS_SQL.DEFINE_COLUMN (c_stat, v_cnt, v_number);
	  END IF;
	END LOOP;

	--dbms_output.put_line ('--Start--');
  LOOP
  	--dbms_output.put_line (v_colcnt);
    IF SYS.DBMS_SQL.FETCH_ROWS (c_stat) >0 THEN
    	--dbms_output.put_line ('innerhalb Fetch');
      for v_cnt in 1..v_colcnt
			LOOP
    		IF (v_coltype(v_cnt)='V') THEN
  	  		  dbms_sql.column_value(c_stat, v_cnt, v_string);
						INSERT INTO oqs_kpi_log_run (kpi_time, kpi_name,
                   kpi_type, kpi_c_val,  kpi_group_val)
									 VALUES (v_starttime, v_colname(v_cnt), 'V',
									        v_string,  group_value)
						;
						COMMIT;
            dbms_output.put_line(v_colname(v_cnt) || ' := '||v_string);
			  END IF;
			  IF (v_coltype(v_cnt)='N') THEN
	  		  dbms_sql.column_value(c_stat, v_cnt, v_number);
						INSERT INTO oqs_kpi_log_run (kpi_time, kpi_name,
                   kpi_type, kpi_n_val, kpi_group_val)
									 VALUES (v_starttime, v_colname(v_cnt), 'N',
									        v_number, group_value)
						;
						COMMIT;
					-- print the name of the KPI and the value
          --dbms_output.put_line(v_colname(v_cnt) || ' := '||v_number);
			  END IF;
			END LOOP;
		END IF;
  END LOOP;

  dbms_sql.close_cursor(c_stat);
	

EXCEPTION
  WHEN CURSOR_CLOSED THEN
    IF SYS.DBMS_SQL.IS_OPEN (c_stat) THEN
      SYS.DBMS_SQL.CLOSE_CURSOR (c_stat);
    END IF;
	WHEN NO_DATA_FOUND THEN
	  dbms_output.put_line ('OQS-00001: Unknown table');
	WHEN STATEMENT_TOO_LONG THEN
	  dbms_output.put_line ('OQS-00002: Statement too long');
  WHEN OTHERS THEN
  	DBMS_OUTPUT.PUT_LINE ('Exception');
    RAISE;

END;
/

show errors;
